/* Test sipe.h against GNU MPFR
 *
 * Usage: sipe-test [ <prec_max> ]
 *
 * Copyright (c) 2011-2012 Vincent Lefevre <vincent@vinc17.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see http://www.gnu.org/licenses/
 * or write to the Free Software Foundation, Inc., 51 Franklin St,
 * Fifth Floor, Boston, MA 02110-1301, USA.
 */

#define SVNID "$Id: sipe-test.c 56953 2012-12-11 15:05:07Z vinc17/ypig $"

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <mpfr.h>

#define SIPE_USE_FMA 1
#include "sipe.h"

#define STRINGIZE(S) #S
#define MAKE_STR(S) STRINGIZE(S)

/* Syntax check of SIPE_ROUND (during compilation). */
static void check_round_macro (int prec)
{
  sipe_t r = { 1, 0 };

  if (1)
    SIPE_ROUND (r, prec);
  else
    SIPE_ROUND (r, prec);
}

sipe_t first (mpfr_ptr u, int prec)
{
  sipe_t r = { 1, -2 * prec };

  mpfr_set_ui_2exp (u, 1, -2 * prec, MPFR_RNDN);
  SIPE_ROUND (r, prec);
  return r;
}

#define LOOP(U,BODY)                                    \
  do                                                    \
    for (U##s = first (U##m, prec);                     \
         U##s.e < 2 * prec;                             \
         mpfr_nextabove (U##m),                         \
           U##s = sipe_nextabove (U##s, prec))          \
      for (U##n = 1; U##n >= 0; U##n--)                 \
        { BODY;                                         \
          mpfr_neg (U##m, U##m, MPFR_RNDN);             \
          U##s = sipe_neg (U##s, prec); }               \
  while (0)

int main (int argc, char **argv)
{
  int prec;
  long prec_max = SIPE_PREC_MAX;
  long ctr_add = 0, ctr_sub = 0, ctr_mul = 0, ctr_fma = 0, ctr_fms = 0,
    ctr_2mul = 0, ctr_add_si = 0, ctr_sub_si = 0, ctr_mul_si = 0,
    ctr_eq = 0, ctr_ne = 0, ctr_le = 0, ctr_gt = 0, ctr_ge = 0, ctr_lt = 0,
    ctr_min = 0, ctr_max = 0;

#include <version-info.h>

  if (argc > 2)
    {
      fprintf (stderr, "Usage: sipe-test [ <prec_max> ]\n");
      exit (1);
    }
  else if (argc == 2)
    {
      char *end;
      prec_max = strtol (argv[1], &end, 10);
      if (*end != '\0' || prec_max < 2 || prec_max > SIPE_PREC_MAX)
        {
          fprintf (stderr, "sipe-test: bad prec_max value\n");
          exit (1);
        }
    }

  for (prec = 2; prec <= prec_max; prec++)
    {
      mpfr_t xm, ym, zm, rm, sm;
      sipe_t xs, ys, zs, rs, ss;
      int xn, yn, zn;
      int bm, bsf, bst;
      sipe_int_t i, imax;
      clock_t t;

      printf ("Testing precision %d...", prec);
      fflush (stdout);
      t = clock ();
      check_round_macro (prec);
      imax = (sipe_int_t) 1 << prec;
      mpfr_inits2 (prec, xm, ym, zm, rm, sm, (mpfr_ptr) 0);
      LOOP (x,
        {
          LOOP (y,
            {
              ++ctr_add;
              rs = sipe_add (xs, ys, prec);
              mpfr_set_si_2exp (sm, rs.i, rs.e, MPFR_RNDN);
              mpfr_add (rm, xm, ym, MPFR_RNDN);
              if (! mpfr_equal_p (rm, sm))
                {
                  putchar ('\n');
                  fflush (stdout);
                  mpfr_fprintf
                    (stderr,
                     "sipe-test: error for add on %Rb / %Rb\n"
                     "expected %Rb\n"
                     "     got %Rb\n",
                     xm, ym, rm, sm);
                  exit (1);
                }
              ++ctr_sub;
              rs = sipe_sub (xs, ys, prec);
              mpfr_set_si_2exp (sm, rs.i, rs.e, MPFR_RNDN);
              mpfr_sub (rm, xm, ym, MPFR_RNDN);
              if (! mpfr_equal_p (rm, sm))
                {
                  putchar ('\n');
                  fflush (stdout);
                  mpfr_fprintf
                    (stderr,
                     "sipe-test: error for sub on %Rb / %Rb\n"
                     "expected %Rb\n"
                     "     got %Rb\n",
                     xm, ym, rm, sm);
                  exit (1);
                }
              ++ctr_mul;
              rs = sipe_mul (xs, ys, prec);
              mpfr_set_si_2exp (sm, rs.i, rs.e, MPFR_RNDN);
              mpfr_mul (rm, xm, ym, MPFR_RNDN);
              if (! mpfr_equal_p (rm, sm))
                {
                  putchar ('\n');
                  fflush (stdout);
                  mpfr_fprintf
                    (stderr,
                     "sipe-test: error for mul on %Rb / %Rb\n"
                     "expected %Rb\n"
                     "     got %Rb\n",
                     xm, ym, rm, sm);
                  exit (1);
                }
              ++ctr_2mul;
              SIPE_2MUL (zs, ss, xs, ys, prec);
              mpfr_set_si_2exp (sm, ss.i, ss.e, MPFR_RNDN);
              mpfr_fms (rm, xm, ym, rm, MPFR_RNDN);
              if (! sipe_eq (rs, zs) || ! mpfr_equal_p (rm, sm))
                {
                  putchar ('\n');
                  fflush (stdout);
                  mpfr_fprintf
                    (stderr, "sipe-test: error for SIPE_2MUL on %Rb / %Rb\n"
                     "expected (", xm, ym);
                  sipe_outbin (stderr, rs, prec);
                  mpfr_fprintf (stderr, ",%Rb)\n     got (", rm);
                  sipe_outbin (stderr, zs, prec);
                  mpfr_fprintf (stderr, ",%Rb)\n", sm);
                  exit (1);
                }
              LOOP (z,
                {
                  ++ctr_fma;
                  rs = sipe_fma (xs, ys, zs, prec);
                  mpfr_set_si_2exp (sm, rs.i, rs.e, MPFR_RNDN);
                  mpfr_fma (rm, xm, ym, zm, MPFR_RNDN);
                  if (! mpfr_equal_p (rm, sm))
                    {
                      putchar ('\n');
                      fflush (stdout);
                      mpfr_fprintf
                        (stderr,
                         "sipe-test: error for fma on %Rb / %Rb / %Rb\n"
                         "expected %Rb\n"
                         "     got %Rb\n",
                         xm, ym, zm, rm, sm);
                      exit (1);
                    }
                  ++ctr_fms;
                  rs = sipe_fms (xs, ys, zs, prec);
                  mpfr_set_si_2exp (sm, rs.i, rs.e, MPFR_RNDN);
                  mpfr_fms (rm, xm, ym, zm, MPFR_RNDN);
                  if (! mpfr_equal_p (rm, sm))
                    {
                      putchar ('\n');
                      fflush (stdout);
                      mpfr_fprintf
                        (stderr,
                         "sipe-test: error for fms on %Rb / %Rb / %Rb\n"
                         "expected %Rb\n"
                         "     got %Rb\n",
                         xm, ym, zm, rm, sm);
                      exit (1);
                    }
                } );  /* LOOP (z, ...) */
              if (prec <= 3)
                {
                  ++ctr_eq;
                  bsf = sipe_eq (xs, ys);
                  ++ctr_ne;
                  bst = sipe_ne (xs, ys);
                  bm = ! mpfr_equal_p (xm, ym);
                  if (bsf == bm || bst != bm)
                    {
                      putchar ('\n');
                      fflush (stdout);
                      mpfr_fprintf
                        (stderr,
                         "sipe-test: error for sipe_eq/ne on %Rb / %Rb\n"
                         "(%d / %d / %d)\n", xm, ym, bm, bsf, bst);
                      exit (1);
                    }
                  ++ctr_le;
                  bsf = sipe_le (xs, ys);
                  ++ctr_gt;
                  bst = sipe_gt (xs, ys);
                  bm = ! mpfr_lessequal_p (xm, ym);
                  if (bsf == bm || bst != bm)
                    {
                      putchar ('\n');
                      fflush (stdout);
                      mpfr_fprintf
                        (stderr,
                         "sipe-test: error for sipe_le/gt on %Rb / %Rb\n"
                         "(%d / %d / %d)\n", xm, ym, bm, bsf, bst);
                      exit (1);
                    }
                  ++ctr_ge;
                  bsf = sipe_ge (xs, ys);
                  ++ctr_lt;
                  bst = sipe_lt (xs, ys);
                  bm = ! mpfr_greaterequal_p (xm, ym);
                  if (bsf == bm || bst != bm)
                    {
                      putchar ('\n');
                      fflush (stdout);
                      mpfr_fprintf
                        (stderr,
                         "sipe-test: error for sipe_ge/lt on %Rb / %Rb\n"
                         "(%d / %d / %d)\n", xm, ym, bm, bsf, bst);
                      exit (1);
                    }
                  ++ctr_min;
                  rs = sipe_min (xs, ys);
                  mpfr_set_si_2exp (sm, rs.i, rs.e, MPFR_RNDN);
                  mpfr_min (rm, xm, ym, MPFR_RNDN);
                  if (! mpfr_equal_p (rm, sm))
                    {
                      putchar ('\n');
                      fflush (stdout);
                      mpfr_fprintf
                        (stderr,
                         "sipe-test: error for min on %Rb / %Rb\n"
                         "expected %Rb\n"
                         "     got %Rb\n",
                         xm, ym, rm, sm);
                      exit (1);
                    }
                  ++ctr_max;
                  rs = sipe_max (xs, ys);
                  mpfr_set_si_2exp (sm, rs.i, rs.e, MPFR_RNDN);
                  mpfr_max (rm, xm, ym, MPFR_RNDN);
                  if (! mpfr_equal_p (rm, sm))
                    {
                      putchar ('\n');
                      fflush (stdout);
                      mpfr_fprintf
                        (stderr,
                         "sipe-test: error for max on %Rb / %Rb\n"
                         "expected %Rb\n"
                         "     got %Rb\n",
                         xm, ym, rm, sm);
                      exit (1);
                    }
                }  /* prec <= 3 */
            } );  /* LOOP (y, ...) */
          for (i = -imax; i <= imax; i++)
            {
              ++ctr_add_si;
              rs = sipe_add_si (xs, i, prec);
              mpfr_set_si_2exp (sm, rs.i, rs.e, MPFR_RNDN);
              mpfr_add_si (rm, xm, i, MPFR_RNDN);
              if (! mpfr_equal_p (rm, sm))
                {
                  putchar ('\n');
                  fflush (stdout);
                  mpfr_fprintf
                    (stderr,
                     "sipe-test: error for add_si on %Rb / %d\n"
                     "expected %Rb\n"
                     "     got %Rb\n",
                     xm, i, rm, sm);
                  exit (1);
                }
              ++ctr_sub_si;
              rs = sipe_sub_si (xs, i, prec);
              mpfr_set_si_2exp (sm, rs.i, rs.e, MPFR_RNDN);
              mpfr_sub_si (rm, xm, i, MPFR_RNDN);
              if (! mpfr_equal_p (rm, sm))
                {
                  putchar ('\n');
                  fflush (stdout);
                  mpfr_fprintf
                    (stderr,
                     "sipe-test: error for sub_si on %Rb / %d\n"
                     "expected %Rb\n"
                     "     got %Rb\n",
                     xm, i, rm, sm);
                  exit (1);
                }
              ++ctr_mul_si;
              rs = sipe_mul_si (xs, i, prec);
              mpfr_set_si_2exp (sm, rs.i, rs.e, MPFR_RNDN);
              mpfr_mul_si (rm, xm, i, MPFR_RNDN);
              if (! mpfr_equal_p (rm, sm))
                {
                  putchar ('\n');
                  fflush (stdout);
                  mpfr_fprintf
                    (stderr,
                     "sipe-test: error for mul_si on %Rb / %d\n"
                     "expected %Rb\n"
                     "     got %Rb\n",
                     xm, i, rm, sm);
                  exit (1);
                }
            }
        } );  /* LOOP (x, ...) */
      mpfr_clears (xm, ym, zm, rm, sm, (mpfr_ptr) 0);
      t = clock () - t;
      printf (" %.2f s\n", (double) t / CLOCKS_PER_SEC);
      fflush (stdout);
    }

  printf ("Number of tests with SIPE_ROUND_ZOPT = " MAKE_STR(SIPE_ROUND_ZOPT)
          " (%ld):\n", (long) SIPE_ROUND_ZOPT);
  printf ("%12ld sipe_add\n", ctr_add);
  printf ("%12ld sipe_sub\n", ctr_sub);
  printf ("%12ld sipe_mul\n", ctr_mul);
  printf ("%12ld SIPE_2MUL\n", ctr_2mul);
  printf ("%12ld sipe_fma\n", ctr_fma);
  printf ("%12ld sipe_fms\n", ctr_fms);
  printf ("%12ld sipe_eq\n", ctr_eq);
  printf ("%12ld sipe_ne\n", ctr_ne);
  printf ("%12ld sipe_le\n", ctr_le);
  printf ("%12ld sipe_gt\n", ctr_gt);
  printf ("%12ld sipe_ge\n", ctr_ge);
  printf ("%12ld sipe_lt\n", ctr_lt);
  printf ("%12ld sipe_min\n", ctr_min);
  printf ("%12ld sipe_max\n", ctr_max);
  printf ("%12ld sipe_add_si\n", ctr_add_si);
  printf ("%12ld sipe_sub_si\n", ctr_sub_si);
  printf ("%12ld sipe_mul_si\n", ctr_mul_si);

  return 0;
}
